﻿/*
 Copyright (c) 2011 Seth Juarez

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
*/

using System;
using ml.Math;
using System.IO;
using System.Xml.Serialization;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace ml.Tests
{
    /// <summary>
    /// Summary description for MatrixTests
    /// </summary>
    [TestClass]
    public class MatrixTests
    {
        [TestMethod]
        public void Matrix_Serialize_Test()
        {
            string path = Directory.GetCurrentDirectory() + @"\matrix_serialize_test.xml";

            Matrix m1 = new[,] {
                { System.Math.PI, System.Math.PI / 2.3, System.Math.PI * 1.2, System.Math.PI, System.Math.PI / 2.3, System.Math.PI * 1.2 },
                { System.Math.PI, System.Math.PI / 2.3, System.Math.PI * 1.2, System.Math.PI, System.Math.PI / 2.3, System.Math.PI * 1.2 },
                { System.Math.PI, System.Math.PI / 2.3, System.Math.PI * 1.2, System.Math.PI, System.Math.PI / 2.3, System.Math.PI * 1.2 },
                { System.Math.PI, System.Math.PI / 2.3, System.Math.PI * 1.2, System.Math.PI, System.Math.PI / 2.3, System.Math.PI * 1.2 },
                { System.Math.PI, System.Math.PI / 2.3, System.Math.PI * 1.2, System.Math.PI, System.Math.PI / 2.3, System.Math.PI * 1.2 }
            };

            XmlSerializer serializer = new XmlSerializer(typeof(Matrix));

            // serialize
            // ensure we delete the file first or we may have extra data
            if (File.Exists(path))
            {
                // this could get an access violation but either way 
                // we don't want a pointer stuck in the app domain
                File.Delete(path);
            }

            using (var stream = File.OpenWrite(path))
            {
                serializer.Serialize(stream, m1);
            }

            // deserialize
            Matrix m2 = null;
            using (var stream = File.OpenRead(path))
            {
                m2 = (Matrix)serializer.Deserialize(stream);
            }

            Assert.AreEqual(m1, m2);
        }

        [TestMethod]
        public void Matrix_Equal_Test()
        {
            Matrix one = new[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}};

            Matrix two = new[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}};

            Assert.AreEqual<bool>(true, one.Equals(two));
            Assert.AreEqual<bool>(true, one == two);
        }

        [TestMethod]
        public void Matrix_Not_Equal_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 10}});

            Matrix two = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            Assert.AreEqual<bool>(false, one.Equals(two));
            Assert.AreEqual<bool>(false, one == two);
            Assert.AreEqual<bool>(true, one != two);
        }

        [TestMethod]
        public void Matrix_Identity_Test()
        {
            Matrix eye1 = new Matrix(new double[,]
                {{1, 0, 0},
                 {0, 1, 0},
                 {0, 0, 1}});

            Matrix eye2 = new Matrix(new double[,]
                {{1, 0, 0, 0},
                 {0, 1, 0, 0},
                 {0, 0, 1, 0}});

            Matrix eye3 = new Matrix(new double[,]
                {{1, 0, 0},
                 {0, 1, 0},
                 {0, 0, 1},
                 {0, 0, 0}});

            Assert.AreEqual<Matrix>(eye1, Matrix.Identity(3));
            Assert.AreEqual<Matrix>(eye2, Matrix.Identity(3, 4));
            Assert.AreEqual<Matrix>(eye3, Matrix.Identity(4, 3));
        }

        [TestMethod]
        public void Matrix_Trace_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 10}});

            Matrix two = new Matrix(new double[,]
                {{1, 2, 3, 9},
                 {4, 5, 6, 12},
                 {7, 8, 10, 13}});

            Matrix three = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 10},
                 {11,12,13}});

            Assert.AreEqual<double>(16, one.Trace());
            Assert.AreEqual<double>(16, two.Trace());
            Assert.AreEqual<double>(16, three.Trace());
        }

        [TestMethod]
        public void Matrix_Zeros_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{0, 0, 0},
                 {0, 0, 0},
                 {0, 0, 0}});

            Matrix two = new Matrix(new double[,]
                {{0, 0, 0, 0},
                 {0, 0, 0, 0},
                 {0, 0, 0, 0}});

            Matrix three = new Matrix(new double[,]
                {{0, 0, 0},
                 {0, 0, 0},
                 {0, 0, 0},
                 {0, 0, 0}});

            Assert.AreEqual<Matrix>(one, Matrix.Zeros(3));
            Assert.AreEqual<Matrix>(two, Matrix.Zeros(3, 4));
            Assert.AreEqual<Matrix>(three, Matrix.Zeros(4, 3));
        }

        [TestMethod]
        public void Matrix_Ones_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{1, 1, 1},
                 {1, 1, 1},
                 {1, 1, 1}});

            Matrix two = new Matrix(new double[,]
                {{1, 1, 1, 1},
                 {1, 1, 1, 1},
                 {1, 1, 1, 1}});

            Matrix three = new Matrix(new double[,]
                {{1, 1, 1},
                 {1, 1, 1},
                 {1, 1, 1},
                 {1, 1, 1}});

            Assert.AreEqual<Matrix>(one, Matrix.Ones(3));
            Assert.AreEqual<Matrix>(two, Matrix.Ones(3, 4));
            Assert.AreEqual<Matrix>(three, Matrix.Ones(4, 3));
        }

        [TestMethod]
        public void Matrix_Transpose_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            Matrix oneT = new Matrix(new double[,]
                {{1, 4, 7},
                 {2, 5, 8},
                 {3, 6, 9}});

            Matrix two = new Matrix(new double[,]
                {{1, 2, 3, 9},
                 {4, 5, 6, 12},
                 {7, 8, 10, 13}});

            Matrix twoT = new Matrix(new double[,]
                {{1, 4, 7},
                 {2, 5, 8},
                 {3, 6, 10},
                 {9,12, 13}});

            Assert.AreEqual<Matrix>(oneT, one.Transpose());
            Assert.AreEqual<Matrix>(oneT, one.T);
            Assert.AreEqual<Matrix>(one, oneT.Transpose());
            Assert.AreEqual<Matrix>(one, oneT.T);
            Assert.AreEqual<Matrix>(twoT, two.Transpose());
            Assert.AreEqual<Matrix>(twoT, two.T);
            Assert.AreEqual<Matrix>(two, twoT.Transpose());
            Assert.AreEqual<Matrix>(two, twoT.T);
        }

        [TestMethod]
        public void Matrix_Assign_Value_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            one[1, 1] = 14.5;
            Assert.AreEqual<double>(14.5, one[1, 1]);
        }

        [TestMethod]
        [ExpectedException(typeof(IndexOutOfRangeException))]
        public void Matrix_Assign_Value_Bad_Index_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            one[5, 5] = 14.5;
        }

        [TestMethod]
        [ExpectedException(typeof(IndexOutOfRangeException))]
        public void Matrix_Read_Value_Bad_Index_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            var d = one[5, 5];
        }

        [TestMethod]
        public void Matrix_Add_Aligned_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            Matrix two = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            Matrix sum = new Matrix(new double[,]
                {{2, 4, 6},
                 {8, 10, 12},
                 {14, 16, 18}});

            Matrix a = new Matrix(new double[,]
                {{1, 2, 3, 1},
                 {4, 5, 6, 2},
                 {7, 8, 9, 3}});

            Matrix b = new Matrix(new double[,]
                {{1, 2, 3, 5},
                 {4, 5, 6, 6},
                 {7, 8, 9, 7}});

            Matrix c = new Matrix(new double[,]
                {{2, 4, 6, 6},
                 {8, 10, 12, 8},
                 {14, 16, 18, 10}});

            Matrix d = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9},
                 {1, 2, 3}});

            Matrix e = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9},
                 {3, 2, 1}});

            Matrix f = new Matrix(new double[,]
                {{2, 4, 6},
                 {8, 10, 12},
                 {14, 16, 18},
                 {4, 4, 4}});

            Assert.AreEqual<Matrix>(sum, one + two);
            Assert.AreEqual<Matrix>(c, a + b);
            Assert.AreEqual<Matrix>(f, d + e);
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void Matrix_Add_Non_Aligned_Test_1()
        {
            Matrix a = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            Matrix b = new Matrix(new double[,]
                {{2, 4, 6, 6},
                 {8, 10, 12, 8},
                 {14, 16, 18, 10}});

            var x = a + b;
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void Matrix_Add_Non_Aligned_Test_2()
        {
            Matrix a = new Matrix(new double[,]
                {{2, 4, 6},
                 {8, 10, 12},
                 {14, 16, 18},
                 {4, 4, 4}});

            Matrix b = new Matrix(new double[,]
                {{2, 4, 6, 6},
                 {8, 10, 12, 8},
                 {14, 16, 18, 10}});

            var x = a + b;
        }

        [TestMethod]
        public void Matrix_Subtract_Aligned_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            Matrix two = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            Matrix sum = new Matrix(new double[,]
                {{0, 0, 0},
                 {0, 0, 0},
                 {0, 0, 0}});

            Matrix a = new Matrix(new double[,]
                {{1, 2, 3, 1},
                 {4, 5, 6, 2},
                 {7, 8, 9, 3}});

            Matrix b = new Matrix(new double[,]
                {{1, 2, 3, 5},
                 {4, 5, 6, 6},
                 {7, 8, 9, 7}});

            Matrix c = new Matrix(new double[,]
                {{0, 0, 0, -4},
                 {0, 0, 0, -4},
                 {0, 0, 0, -4}});

            Matrix d = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9},
                 {1, 2, 3}});

            Matrix e = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9},
                 {3, 2, 1}});

            Matrix f = new Matrix(new double[,]
                {{0, 0, 0},
                 {0, 0, 0},
                 {0, 0, 0},
                 {-2, 0, 2}});

            Assert.AreEqual<Matrix>(sum, one - two);
            Assert.AreEqual<Matrix>(c, a - b);
            Assert.AreEqual<Matrix>(f, d - e);
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void Matrix_Subtract_Non_Aligned_Test_1()
        {
            Matrix a = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            Matrix b = new Matrix(new double[,]
                {{2, 4, 6, 6},
                 {8, 10, 12, 8},
                 {14, 16, 18, 10}});

            var x = a - b;
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void Matrix_Subtract_Non_Aligned_Test_2()
        {
            Matrix a = new Matrix(new double[,]
                {{2, 4, 6},
                 {8, 10, 12},
                 {14, 16, 18},
                 {4, 4, 4}});

            Matrix b = new Matrix(new double[,]
                {{2, 4, 6, 6},
                 {8, 10, 12, 8},
                 {14, 16, 18, 10}});

            var x = a - b;
        }

        [TestMethod]
        public void Matrix_Multiply_Vector_Aligned_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{ 2,  4,  6},
                 { 8, 10, 12},
                 {14, 16, 18}});
            Vector v = new Vector(new double[] { 0.5, 0.5, 0.5 });

            Matrix s1 = (new Vector(new double[] { 6, 15, 24 }))
                            .ToMatrix(VectorType.Column);
            Matrix s2 = (new Vector(new double[] { 12, 15, 18 }))
                            .ToMatrix(VectorType.Row);

            Assert.AreEqual<Matrix>(s1, one * v);
            Assert.AreEqual<Matrix>(s2, v * one);
        }

        [TestMethod]
        public void Matrix_Sum_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{1, 2, 3, 4},
                 {4, 5, 6, 5},
                 {7, 8, 9, 6}});

            Vector row = new Vector(new double[] { 12, 15, 18, 15 });
            Vector col = new Vector(new double[] { 10, 20, 30 });

            Assert.AreEqual<Vector>(row, one.Sum(VectorType.Row));
            Assert.AreEqual<Vector>(col, one.Sum(VectorType.Column));
            Assert.AreEqual<double>(60, one.Sum());
        }

        [TestMethod]
        public void Matrix_Multiply_Aligned_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            Matrix sol = new Matrix(new double[,]
                {{30, 36, 42},
                 {66, 81, 96},
                 {102, 126, 150}});

            Matrix a = new Matrix(new double[,]
                {{2, 4, 6},
                 {8, 10, 12},
                 {14, 16, 18},
                 {4, 4, 4}});

            Matrix b = new Matrix(new double[,]
                {{2, 4, 6, 6},
                 {8, 10, 12, 8},
                 {14, 16, 18, 10}});

            Matrix c = new Matrix(new double[,]
                {{120, 144, 168, 104},
                 {264, 324, 384, 248},
                 {408, 504, 600, 392},
                 { 96, 120, 144,  96}});


            Assert.AreEqual<Matrix>(sol, one * one);
            Assert.AreEqual<Matrix>(c, a * b);
        }

        [TestMethod]
        public void Matrix_Multiply_Scalar_Test()
        {
            Matrix one = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            Matrix sol = new Matrix(new double[,]
                {{ 2,  4,  6},
                 { 8, 10, 12},
                 {14, 16, 18}});

            Assert.AreEqual<Matrix>(sol, 2 * one);
            Assert.AreEqual<Matrix>(sol, one * 2);
        }

        [TestMethod]
        [ExpectedException(typeof(InvalidOperationException))]
        public void Matrix_Multiply_Non_Aligned_Test()
        {
            Matrix a = new Matrix(new double[,]
                {{1, 2, 3},
                 {4, 5, 6},
                 {7, 8, 9}});

            Matrix b = new Matrix(new double[,]
                {{2, 4, 6},
                 {8, 10, 12},
                 {14, 16, 18},
                 {4, 4, 4}});

            var x = a * b;
        }

        [TestMethod]
        public void Matrix_Inverse_Test()
        {
            Matrix a = new Matrix(new double[,]
                {{4, 3},
                 {3, 2}});

            Matrix aInv = new Matrix(new double[,]
                {{-2,  3},
                 { 3, -4}});

            Matrix b = new Matrix(new double[,]
                {{ 1,  2,  3},
                 { 0,  4,  5},
                 { 1,  0,  6}});

            Matrix bInv = new Matrix(new double[,]
                {{ 12.0/11,  -6.0/11,  -1.0/11},
                 {  5.0/22,   3.0/22,  -5.0/22},
                 { -2.0/11,   1.0/11,   2.0/11}});

            Assert.AreEqual<Matrix>(aInv, a ^ -1);
            Assert.AreEqual<string>(bInv.ToString(), (b ^ -1).ToString());
        }


        [TestMethod]
        public void Matrix_Cholesky_Test()
        {
            Matrix m = new Matrix(new double[,]
                {{ 25, 15, -5},
                 { 15, 18,  0},
                 { -5,  0, 11}});

            Matrix sol = new Matrix(new double[,]
                {{  5, 0, 0},
                 {  3, 3, 0},
                 { -1, 1, 3}});

            Assert.AreEqual(sol, Matrix.Cholesky(m));
            Assert.AreEqual(sol, m.Cholesky());
        }

        [TestMethod]
        public void Matrix_LLS_Test()
        {
            Matrix A = new Matrix(new double[,]
                {{ 3, -6},
                 { 4, -8},
                 { 0,  1}});

            Vector b = new Vector(new double[] { -1, 7, 2 });
            Vector sol = new Vector(new double[] { 5, 2 });

            // LLS implementation
            Vector x = A / b;
            Assert.AreEqual(sol, x);

            // this should fire regular inverse
            // need to be a bit smarter about using straight
            // inverse here...
            // Cholesky if positive diagonal or
            // LU
            Vector y = (A.T * A) / (A.T * b).ToVector();
            Assert.AreEqual(sol, y);
        }

        [TestMethod]
        public void Matrix_QR_Test()
        {
            Matrix A = new[,] {{ 4,  1,  2},
                                 { 1,  4,  0},
                                 { 2,  0,  4}};

            var t = Matrix.QR(A);
            var Q = t.Item1;
            var R = t.Item2;

            // close enough...
            var diff = A.Norm() - (Q * R).Norm();
            Assert.AreEqual(0, diff);
        }

        [TestMethod]
        public void Matrix_Eigs_Test()
        {
            Matrix A = new Matrix(new double[,]
                {{  5,  2,  4 },
                 { -3,  6,  2 },
                 {  3, -3,  1 }});

            var eigenValues = new[] { 7, 2, 3 };
            var eigs = A.Eigs();
            var S = eigs.Item1;
            var V = eigs.Item2;

            //Assert.AreEqual(eigenValues.Round(), S.Round());

        }

        [TestMethod]
        public void Matrix_SVD_Test()
        {
            Matrix A = new[,]
                {{  5,  2,  4 },
                 { -3,  6,  2 },
                 {  3, -3,  1 }};

            var svd = A.SVD();
            var U = svd.Item1;
            var S = svd.Item2;
            var V = svd.Item3;

            Assert.AreEqual(A.Round(8), (U * S.Diag() * V.T).Round(8));
        }


        [TestMethod]
        public void Matrix_Extract_Test()
        {
            Matrix A = new [,]
                {{  1,  2,  4 },
                 {  1,  6,  2 },
                 {  3,  1,  1 }};


            Matrix sol = new[,]
                {{ 6, 2},
                 { 1, 1}};

            Assert.AreEqual(sol, A.Extract(1, 1, 2, 2));

            Matrix B = new[,]
                {{  1,  2,  4,  9 },
                 {  1,  6,  2,  7 },
                 {  3,  1,  1,  4 },
                 {  4,  2,  3,  2 }};

            Matrix bSol = new[,]
                {{  2,  7 },
                 {  1,  4 },
                 {  3,  2 }};

            Assert.AreEqual(bSol, B.Extract(2, 1, 2, 3));
        }
    }
}
